@@ -25,18 +25,20 @@
 //
 
 use std::fmt;
+use std::fmt::Write as _;
 
-use endian::Endian;
+use crate::endian::Endian;
+use crate::error::Error;
 
-/// Types and values of TIFF fields (for Exif attributes).
-#[derive(Debug)]
-pub enum Value<'a> {
+/// A type and values of a TIFF/Exif field.
+#[derive(Clone)]
+pub enum Value {
     /// Vector of 8-bit unsigned integers.
     Byte(Vec<u8>),
     /// Vector of slices of 8-bit bytes containing 7-bit ASCII characters.
     /// The trailing null characters are not included.  Note that
     /// the 8th bits may present if a non-conforming data is given.
-    Ascii(Vec<&'a [u8]>),
+    Ascii(Vec<Vec<u8>>),
     /// Vector of 16-bit unsigned integers.
     Short(Vec<u16>),
     /// Vector of 32-bit unsigned integers.
@@ -52,7 +54,7 @@ pub enum Value<'a> {
     /// The interpretation of the value does not generally depend on
     /// the location, but if it does, the offset information helps.
     /// When encoding Exif, it is ignored.
-    Undefined(&'a [u8], u32),
+    Undefined(Vec<u8>, u32),
     /// Vector of 16-bit signed integers.  Unused in the Exif specification.
     SShort(Vec<i16>),
     /// Vector of 32-bit signed integers.
@@ -72,25 +74,85 @@ pub enum Value<'a> {
     Unknown(u16, u32, u32),
 }
 
-impl<'a> Value<'a> {
+impl Value {
     /// Returns an object that implements `std::fmt::Display` for
     /// printing a value in a tag-specific format.
     /// The tag of the value is specified as the argument.
     ///
+    /// If you want to display with the unit, use `Field::display_value`.
+    ///
     /// # Examples
     ///
     /// ```
     /// use exif::{Value, Tag};
-    /// let val = Value::Undefined(b"0231", 0);
-    /// assert_eq!(format!("{}", val.display_as(Tag::ExifVersion)),
-    ///            "2.31");
+    /// let val = Value::Undefined(b"0231".to_vec(), 0);
+    /// assert_eq!(val.display_as(Tag::ExifVersion).to_string(), "2.31");
     /// let val = Value::Short(vec![2]);
-    /// assert_eq!(format!("{}", val.display_as(Tag::ResolutionUnit)),
-    ///            "pixels per inch");
+    /// assert_eq!(val.display_as(Tag::ResolutionUnit).to_string(), "inch");
     /// ```
     #[inline]
-    pub fn display_as(&self, tag: ::tag_priv::Tag) -> Display {
-        ::tag_priv::display_value_as(self, tag)
+    pub fn display_as(&self, tag: crate::tag::Tag) -> Display {
+        crate::tag::display_value_as(self, tag)
+    }
+
+    /// Returns the value as a slice if the type is BYTE.
+    #[inline]
+    pub(crate) fn byte(&self) -> Option<&[u8]> {
+        match *self {
+            Value::Byte(ref v) => Some(v),
+            _ => None,
+        }
+    }
+
+    /// Returns the value as `AsciiValues` if the type is ASCII.
+    #[inline]
+    pub(crate) fn ascii(&self) -> Option<AsciiValues> {
+        match *self {
+            Value::Ascii(ref v) => Some(AsciiValues(v)),
+            _ => None,
+        }
+    }
+
+    /// Returns the value as a slice if the type is RATIONAL.
+    #[inline]
+    pub(crate) fn rational(&self) -> Option<&[Rational]> {
+        match *self {
+            Value::Rational(ref v) => Some(v),
+            _ => None,
+        }
+    }
+
+    /// Returns the value as a slice if the type is UNDEFINED.
+    #[inline]
+    pub(crate) fn undefined(&self) -> Option<&[u8]> {
+        match *self {
+            Value::Undefined(ref v, _) => Some(v),
+            _ => None,
+        }
+    }
+
+    /// Returns `UIntValue` if the value type is unsigned integer (BYTE,
+    /// SHORT, or LONG).  Otherwise `exif::Error` is returned.
+    ///
+    /// The integer(s) can be obtained by `get(&self, index: usize)` method
+    /// on `UIntValue`, which returns `Option<u32>`.
+    /// `None` is returned if the index is out of bounds.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use exif::Value;
+    /// # fn main() -> std::result::Result<(), Box<dyn std::error::Error>> {
+    /// let v = Value::Byte(vec![1u8, 2]);
+    /// assert_eq!(v.as_uint()?.get(0), Some(1u32));
+    /// assert_eq!(v.as_uint()?.get(2), None);
+    /// let v = Value::SLong(vec![1, 2]);
+    /// assert!(v.as_uint().is_err());
+    /// # Ok(()) }
+    /// ```
+    #[inline]
+    pub fn as_uint(&self) -> Result<&UIntValue, Error> {
+        UIntValue::ref_from(self)
     }
 
     /// Returns the unsigned integer at the given position.
@@ -124,9 +186,42 @@ impl<'a> Value<'a> {
     }
 }
 
+pub struct AsciiValues<'a>(&'a [Vec<u8>]);
+
+impl<'a> AsciiValues<'a> {
+    pub fn first(&self) -> Option<&'a [u8]> {
+        self.0.first().map(|x| &x[..])
+    }
+}
+
+#[derive(Debug)]
+#[repr(transparent)]
+pub struct UIntValue(Value);
+
+impl UIntValue {
+    #[inline]
+    fn ref_from(v: &Value) -> Result<&Self, Error> {
+        match *v {
+            Value::Byte(_) | Value::Short(_) | Value::Long(_) =>
+                Ok(unsafe { &*(v as *const Value as *const Self) }),
+            _ => Err(Error::UnexpectedValue("Not unsigned integer")),
+        }
+    }
+
+    #[inline]
+    pub fn get(&self, index: usize) -> Option<u32> {
+        match self.0 {
+            Value::Byte(ref v) => v.get(index).map(|&x| x.into()),
+            Value::Short(ref v) => v.get(index).map(|&x| x.into()),
+            Value::Long(ref v) => v.get(index).map(|&x| x),
+            _ => panic!(),
+        }
+    }
+}
+
 // A struct that wraps std::slice::Iter<'a, u8/u16/u32>.
 pub struct UIntIter<'a> {
-    iter: Box<ExactSizeIterator<Item=u32> + 'a>
+    iter: Box<dyn ExactSizeIterator<Item=u32> + 'a>
 }
 
 impl<'a> Iterator for UIntIter<'a> {
@@ -146,9 +241,10 @@ impl<'a> Iterator for UIntIter<'a> {
 impl<'a> ExactSizeIterator for UIntIter<'a> {}
 
 /// Helper struct for printing a value in a tag-specific format.
+#[derive(Copy, Clone)]
 pub struct Display<'a> {
-    pub fmt: fn(&mut fmt::Write, &Value) -> fmt::Result,
-    pub value: &'a Value<'a>,
+    pub fmt: fn(&mut dyn fmt::Write, &Value) -> fmt::Result,
+    pub value: &'a Value,
 }
 
 impl<'a> fmt::Display for Display<'a> {
@@ -158,6 +254,65 @@ impl<'a> fmt::Display for Display<'a> {
     }
 }
 
+impl fmt::Debug for Value {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::Byte(v) => f.debug_tuple("Byte").field(v).finish(),
+            Self::Ascii(v) => f.debug_tuple("Ascii")
+                .field(&IterDebugAdapter(
+                    || v.iter().map(|x| AsciiDebugAdapter(x)))).finish(),
+            Self::Short(v) => f.debug_tuple("Short").field(v).finish(),
+            Self::Long(v) => f.debug_tuple("Long").field(v).finish(),
+            Self::Rational(v) => f.debug_tuple("Rational").field(v).finish(),
+            Self::SByte(v) => f.debug_tuple("SByte").field(v).finish(),
+            Self::Undefined(v, o) => f.debug_tuple("Undefined")
+                .field(&HexDebugAdapter(v))
+                .field(&format_args!("ofs={:#x}", o)).finish(),
+            Self::SShort(v) => f.debug_tuple("SShort").field(v).finish(),
+            Self::SLong(v) => f.debug_tuple("SLong").field(v).finish(),
+            Self::SRational(v) => f.debug_tuple("SRational").field(v).finish(),
+            Self::Float(v) => f.debug_tuple("Float").field(v).finish(),
+            Self::Double(v) => f.debug_tuple("Double").field(v).finish(),
+            Self::Unknown(t, c, oo) => f.debug_tuple("Unknown")
+                .field(&format_args!("typ={}", t))
+                .field(&format_args!("cnt={}", c))
+                .field(&format_args!("ofs={:#x}", oo)).finish(),
+        }
+    }
+}
+
+struct IterDebugAdapter<F>(F);
+
+impl<F, T, I> fmt::Debug for IterDebugAdapter<F>
+where F: Fn() -> T, T: Iterator<Item = I>, I: fmt::Debug {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.debug_list().entries(self.0()).finish()
+    }
+}
+
+struct AsciiDebugAdapter<'a>(&'a [u8]);
+
+impl<'a> fmt::Debug for AsciiDebugAdapter<'a> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_char('"')?;
+        self.0.iter().try_for_each(|&c| match c {
+            b'\\' | b'"' => write!(f, "\\{}", c as char),
+            0x20..=0x7e => f.write_char(c as char),
+            _ => write!(f, "\\x{:02x}", c),
+        })?;
+        f.write_char('"')
+    }
+}
+
+struct HexDebugAdapter<'a>(&'a [u8]);
+
+impl<'a> fmt::Debug for HexDebugAdapter<'a> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("0x")?;
+        self.0.iter().try_for_each(|x| write!(f, "{:02x}", x))
+    }
+}
+
 // Static default values.
 pub enum DefaultValue {
     None,
@@ -172,16 +327,18 @@ pub enum DefaultValue {
     Unspecified,
 }
 
-impl<'a> From<&'a DefaultValue> for Option<Value<'a>> {
+impl From<&DefaultValue> for Option<Value> {
     fn from(defval: &DefaultValue) -> Option<Value> {
         match *defval {
             DefaultValue::None => None,
             DefaultValue::Byte(s) => Some(Value::Byte(s.to_vec())),
-            DefaultValue::Ascii(s) => Some(Value::Ascii(s.to_vec())),
+            DefaultValue::Ascii(s) => Some(Value::Ascii(
+                s.iter().map(|&x| x.to_vec()).collect())),
             DefaultValue::Short(s) => Some(Value::Short(s.to_vec())),
             DefaultValue::Rational(s) => Some(Value::Rational(
-                s.iter().map(|&t| tuple2rational(t)).collect())),
-            DefaultValue::Undefined(s) => Some(Value::Undefined(s, 0)),
+                s.iter().map(|&x| x.into()).collect())),
+            DefaultValue::Undefined(s) => Some(Value::Undefined(
+                s.to_vec(), 0)),
             DefaultValue::ContextDependent => None,
             DefaultValue::Unspecified => None,
         }
@@ -193,6 +350,12 @@ impl<'a> From<&'a DefaultValue> for Option<Value<'a>> {
 pub struct Rational { pub num: u32, pub denom: u32 }
 
 impl Rational {
+    /// Converts the value to an f32.
+    #[inline]
+    pub fn to_f32(&self) -> f32 {
+        self.to_f64() as f32
+    }
+
     /// Converts the value to an f64.
     #[inline]
     pub fn to_f64(&self) -> f64 {
@@ -200,8 +363,10 @@ impl Rational {
     }
 }
 
-fn tuple2rational(t: (u32, u32)) -> Rational {
-    Rational { num: t.0, denom: t.1 }
+impl From<(u32, u32)> for Rational {
+    fn from(t: (u32, u32)) -> Rational {
+        Rational { num: t.0, denom: t.1 }
+    }
 }
 
 impl fmt::Debug for Rational {
@@ -218,14 +383,16 @@ impl fmt::Display for Rational {
     }
 }
 
+// This implementation has been deprecated.  Use Rational::to_f64 instead.
 impl From<Rational> for f64 {
     #[inline]
     fn from(r: Rational) -> f64 { r.to_f64() }
 }
 
+// This implementation has been deprecated.  Use Rational::to_f32 instead.
 impl From<Rational> for f32 {
     #[inline]
-    fn from(r: Rational) -> f32 { r.to_f64() as f32 }
+    fn from(r: Rational) -> f32 { r.to_f32() }
 }
 
 /// A signed rational number, which is a pair of 32-bit signed integers.
@@ -233,6 +400,12 @@ impl From<Rational> for f32 {
 pub struct SRational { pub num: i32, pub denom: i32 }
 
 impl SRational {
+    /// Converts the value to an f32.
+    #[inline]
+    pub fn to_f32(&self) -> f32 {
+        self.to_f64() as f32
+    }
+
     /// Converts the value to an f64.
     #[inline]
     pub fn to_f64(&self) -> f64 {
@@ -240,6 +413,12 @@ impl SRational {
     }
 }
 
+impl From<(i32, i32)> for SRational {
+    fn from(t: (i32, i32)) -> SRational {
+        SRational { num: t.0, denom: t.1 }
+    }
+}
+
 impl fmt::Debug for SRational {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         write!(f, "SRational({}/{})", self.num, self.denom)
@@ -255,14 +434,16 @@ impl fmt::Display for SRational {
     }
 }
 
+// This implementation has been deprecated.  Use SRational::to_f64 instead.
 impl From<SRational> for f64 {
     #[inline]
     fn from(r: SRational) -> f64 { r.to_f64() }
 }
 
+// This implementation has been deprecated.  Use SRational::to_f32 instead.
 impl From<SRational> for f32 {
     #[inline]
-    fn from(r: SRational) -> f32 { r.to_f64() as f32 }
+    fn from(r: SRational) -> f32 { r.to_f32() }
 }
 
 // Only u32 or i32 are expected for T.
@@ -286,11 +467,10 @@ fn fmt_rational_sub<T>(f: &mut fmt::Formatter, num: u32, denom: T)
     }
 }
 
-type Parser<'a> = fn(&'a [u8], usize, usize) -> Value<'a>;
+type Parser = fn(&[u8], usize, usize) -> Value;
 
 // Return the length of a single value and the parser of the type.
-pub fn get_type_info<'a, E>(typecode: u16)
-                            -> (usize, Parser<'a>) where E: Endian {
+pub fn get_type_info<E>(typecode: u16) -> (usize, Parser) where E: Endian {
     match typecode {
         1 => (1, parse_byte),
         2 => (1, parse_ascii),
@@ -308,25 +488,23 @@ pub fn get_type_info<'a, E>(typecode: u16)
     }
 }
 
-fn parse_byte<'a>(data: &'a [u8], offset: usize, count: usize)
-                  -> Value<'a> {
+fn parse_byte(data: &[u8], offset: usize, count: usize) -> Value {
     Value::Byte(data[offset .. offset + count].to_vec())
 }
 
-fn parse_ascii<'a>(data: &'a [u8], offset: usize, count: usize)
-                   -> Value<'a> {
+fn parse_ascii(data: &[u8], offset: usize, count: usize) -> Value {
     // Any ASCII field can contain multiple strings [TIFF6 Image File
     // Directory].
     let iter = (&data[offset .. offset + count]).split(|&b| b == b'\0');
-    let mut v: Vec<&[u8]> = iter.collect();
-    if v.last().map_or(false, |&s| s.len() == 0) {
+    let mut v: Vec<Vec<u8>> = iter.map(|x| x.to_vec()).collect();
+    if v.last().map_or(false, |x| x.len() == 0) {
         v.pop();
     }
     Value::Ascii(v)
 }
 
-fn parse_short<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                      -> Value<'a> where E: Endian {
+fn parse_short<E>(data: &[u8], offset: usize, count: usize)
+                  -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(E::loadu16(data, offset + i * 2));
@@ -334,8 +512,8 @@ fn parse_short<'a, E>(data: &'a [u8], offset: usize, count: usize)
     Value::Short(val)
 }
 
-fn parse_long<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                     -> Value<'a> where E: Endian {
+fn parse_long<E>(data: &[u8], offset: usize, count: usize)
+                 -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(E::loadu32(data, offset + i * 4));
@@ -343,8 +521,8 @@ fn parse_long<'a, E>(data: &'a [u8], offset: usize, count: usize)
     Value::Long(val)
 }
 
-fn parse_rational<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                         -> Value<'a> where E: Endian {
+fn parse_rational<E>(data: &[u8], offset: usize, count: usize)
+                     -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(Rational {
@@ -355,21 +533,18 @@ fn parse_rational<'a, E>(data: &'a [u8], offset: usize, count: usize)
     Value::Rational(val)
 }
 
-fn parse_sbyte<'a>(data: &'a [u8], offset: usize, count: usize)
-                   -> Value<'a> {
-    let uslice = &data[offset .. offset + count];
-    let islice = unsafe { ::std::slice::from_raw_parts(
-        uslice.as_ptr() as *const i8, count) };
-    Value::SByte(islice.to_vec())
+fn parse_sbyte(data: &[u8], offset: usize, count: usize) -> Value {
+    let bytes = data[offset .. offset + count].iter()
+        .map(|x| *x as i8).collect();
+    Value::SByte(bytes)
 }
 
-fn parse_undefined<'a>(data: &'a [u8], offset: usize, count: usize)
-                       -> Value<'a> {
-    Value::Undefined(&data[offset .. offset + count], offset as u32)
+fn parse_undefined(data: &[u8], offset: usize, count: usize) -> Value {
+    Value::Undefined(data[offset .. offset + count].to_vec(), offset as u32)
 }
 
-fn parse_sshort<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                       -> Value<'a> where E: Endian {
+fn parse_sshort<E>(data: &[u8], offset: usize, count: usize)
+                   -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(E::loadu16(data, offset + i * 2) as i16);
@@ -377,8 +552,8 @@ fn parse_sshort<'a, E>(data: &'a [u8], offset: usize, count: usize)
     Value::SShort(val)
 }
 
-fn parse_slong<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                      -> Value<'a> where E: Endian {
+fn parse_slong<E>(data: &[u8], offset: usize, count: usize)
+                  -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(E::loadu32(data, offset + i * 4) as i32);
@@ -386,8 +561,8 @@ fn parse_slong<'a, E>(data: &'a [u8], offset: usize, count: usize)
     Value::SLong(val)
 }
 
-fn parse_srational<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                          -> Value<'a> where E: Endian {
+fn parse_srational<E>(data: &[u8], offset: usize, count: usize)
+                      -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(SRational {
@@ -399,8 +574,8 @@ fn parse_srational<'a, E>(data: &'a [u8], offset: usize, count: usize)
 }
 
 // TIFF and Rust use IEEE 754 format, so no conversion is required.
-fn parse_float<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                      -> Value<'a> where E: Endian {
+fn parse_float<E>(data: &[u8], offset: usize, count: usize)
+                  -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(f32::from_bits(E::loadu32(data, offset + i * 4)));
@@ -409,8 +584,8 @@ fn parse_float<'a, E>(data: &'a [u8], offset: usize, count: usize)
 }
 
 // TIFF and Rust use IEEE 754 format, so no conversion is required.
-fn parse_double<'a, E>(data: &'a [u8], offset: usize, count: usize)
-                       -> Value<'a> where E: Endian {
+fn parse_double<E>(data: &[u8], offset: usize, count: usize)
+                   -> Value where E: Endian {
     let mut val = Vec::with_capacity(count);
     for i in 0..count {
         val.push(f64::from_bits(E::loadu64(data, offset + i * 8)));
@@ -420,14 +595,13 @@ fn parse_double<'a, E>(data: &'a [u8], offset: usize, count: usize)
 
 // This is a dummy function and will never be called.
 #[allow(unused_variables)]
-fn parse_unknown<'a>(data: &'a [u8], offset: usize, count: usize)
-                     -> Value<'a> {
+fn parse_unknown(data: &[u8], offset: usize, count: usize) -> Value {
     unreachable!()
 }
 
 #[cfg(test)]
 mod tests {
-    use endian::BigEndian;
+    use crate::endian::BigEndian;
     use super::*;
 
     #[test]
@@ -506,8 +680,8 @@ mod tests {
             (b"x", vec![]),
             (b"x\xa1\x02\x03\x04\x05\x06\x07\x08\
                \x09\x0a\x0b\x0c\xbd\x0e\x0f\x10",
-             vec![Rational { num: 0xa1020304, denom: 0x05060708 },
-                  Rational { num: 0x090a0b0c, denom: 0xbd0e0f10 }]),
+             vec![(0xa1020304, 0x05060708).into(),
+                  (0x090a0b0c, 0xbd0e0f10).into()]),
         ];
         let (unitlen, parser) = get_type_info::<BigEndian>(5);
         for &(data, ref ans) in sets {
@@ -598,8 +772,8 @@ mod tests {
             (b"x", vec![]),
             (b"x\xa1\x02\x03\x04\x05\x06\x07\x08\
                \x09\x0a\x0b\x0c\xbd\x0e\x0f\x10",
-             vec![SRational { num: -0x5efdfcfc, denom: 0x05060708 },
-                  SRational { num: 0x090a0b0c, denom: -0x42f1f0f0 }]),
+             vec![(-0x5efdfcfc, 0x05060708).into(),
+                  (0x090a0b0c, -0x42f1f0f0).into()]),
         ];
         let (unitlen, parser) = get_type_info::<BigEndian>(10);
         for &(data, ref ans) in sets {
@@ -621,7 +795,7 @@ mod tests {
         let sets: &[(&[u8], Vec<f32>)] = &[
             (b"x", vec![]),
             (b"x\x7f\x7f\xff\xff\x80\x80\x00\x00\x40\x00\x00\x00",
-             vec![::std::f32::MAX, -::std::f32::MIN_POSITIVE, 2.0]),
+             vec![std::f32::MAX, -std::f32::MIN_POSITIVE, 2.0]),
         ];
         let (unitlen, parser) = get_type_info::<BigEndian>(11);
         for &(data, ref ans) in sets {
@@ -640,7 +814,7 @@ mod tests {
             (b"x\x7f\xef\xff\xff\xff\xff\xff\xff\
                \x80\x10\x00\x00\x00\x00\x00\x00\
                \x40\x00\x00\x00\x00\x00\x00\x00",
-             vec![::std::f64::MAX, -::std::f64::MIN_POSITIVE, 2.0]),
+             vec![std::f64::MAX, -std::f64::MIN_POSITIVE, 2.0]),
         ];
         let (unitlen, parser) = get_type_info::<BigEndian>(12);
         for &(data, ref ans) in sets {
@@ -666,6 +840,24 @@ mod tests {
         assert_eq!(unitlen, 0);
     }
 
+    #[test]
+    fn as_uint() {
+        let v = Value::Byte(vec![1, 2]);
+        assert_eq!(v.as_uint().unwrap().get(0), Some(1));
+        assert_eq!(v.as_uint().unwrap().get(1), Some(2));
+        assert_eq!(v.as_uint().unwrap().get(2), None);
+        let v = Value::Short(vec![1, 2]);
+        assert_eq!(v.as_uint().unwrap().get(0), Some(1));
+        assert_eq!(v.as_uint().unwrap().get(1), Some(2));
+        assert_eq!(v.as_uint().unwrap().get(2), None);
+        let v = Value::Long(vec![1, 2]);
+        assert_eq!(v.as_uint().unwrap().get(0), Some(1));
+        assert_eq!(v.as_uint().unwrap().get(1), Some(2));
+        assert_eq!(v.as_uint().unwrap().get(2), None);
+        let v = Value::SLong(vec![1, 2]);
+        assert_err_pat!(v.as_uint(), Error::UnexpectedValue(_));
+    }
+
     #[test]
     fn get_uint() {
         let v = Value::Byte(vec![1, 2]);
@@ -713,12 +905,44 @@ mod tests {
         assert_eq!(it.len(), 2);
     }
 
+    #[test]
+    fn value_fmt_debug() {
+        let v = Value::Byte(b"b\0y".to_vec());
+        assert_eq!(format!("{:?}", v), "Byte([98, 0, 121])");
+        let v = Value::Ascii(vec![]);
+        assert_eq!(format!("{:?}", v), "Ascii([])");
+        let v = Value::Ascii(vec![b"abc\"\\\n\x7f".to_vec(), b"".to_vec()]);
+        assert_eq!(format!("{:?}", v), r#"Ascii(["abc\"\\\x0a\x7f", ""])"#);
+        let v = Value::Short(vec![]);
+        assert_eq!(format!("{:?}", v), "Short([])");
+        let v = Value::Long(vec![1, 2]);
+        assert_eq!(format!("{:?}", v), "Long([1, 2])");
+        let v = Value::Rational(vec![(0, 0).into()]);
+        assert_eq!(format!("{:?}", v), "Rational([Rational(0/0)])");
+        let v = Value::SByte(vec![-3, 4, 5]);
+        assert_eq!(format!("{:?}", v), "SByte([-3, 4, 5])");
+        let v = Value::Undefined(vec![0, 0xff], 0);
+        assert_eq!(format!("{:?}", v), "Undefined(0x00ff, ofs=0x0)");
+        let v = Value::SShort(vec![6, -7]);
+        assert_eq!(format!("{:?}", v), "SShort([6, -7])");
+        let v = Value::SLong(vec![-9]);
+        assert_eq!(format!("{:?}", v), "SLong([-9])");
+        let v = Value::SRational(vec![(-2, -1).into()]);
+        assert_eq!(format!("{:?}", v), "SRational([SRational(-2/-1)])");
+        let v = Value::Float(vec![1.5, 0.0]);
+        assert_eq!(format!("{:?}", v), "Float([1.5, 0.0])");
+        let v = Value::Double(vec![-0.5, 1.0]);
+        assert_eq!(format!("{:?}", v), "Double([-0.5, 1.0])");
+        let v = Value::Unknown(1, 2, 10);
+        assert_eq!(format!("{:?}", v), "Unknown(typ=1, cnt=2, ofs=0xa)");
+    }
+
     #[test]
     fn rational_fmt_display() {
-        let r = Rational { num: u32::max_value(), denom: u32::max_value() };
+        let r = Rational::from((u32::max_value(), u32::max_value()));
         assert_eq!(format!("{}", r), "4294967295/4294967295");
 
-        let r = Rational { num: 10, denom: 20 };
+        let r = Rational::from((10, 20));
         assert_eq!(format!("{}", r),         "10/20");
         assert_eq!(format!("{:11}", r),      "      10/20");
         assert_eq!(format!("{:3}", r),       "10/20");
@@ -726,22 +950,22 @@ mod tests {
 
     #[test]
     fn srational_fmt_display() {
-        let r = SRational { num: i32::min_value(), denom: i32::min_value() };
+        let r = SRational::from((i32::min_value(), i32::min_value()));
         assert_eq!(format!("{}", r), "-2147483648/-2147483648");
-        let r = SRational { num: i32::max_value(), denom: i32::max_value() };
+        let r = SRational::from((i32::max_value(), i32::max_value()));
         assert_eq!(format!("{}", r), "2147483647/2147483647");
 
-        let r = SRational { num: -10, denom: 20 };
+        let r = SRational::from((-10, 20));
         assert_eq!(format!("{}", r),         "-10/20");
         assert_eq!(format!("{:11}", r),      "     -10/20");
         assert_eq!(format!("{:3}", r),       "-10/20");
 
-        let r = SRational { num: 10, denom: -20 };
+        let r = SRational::from((10, -20));
         assert_eq!(format!("{}", r),         "10/-20");
         assert_eq!(format!("{:11}", r),      "     10/-20");
         assert_eq!(format!("{:3}", r),       "10/-20");
 
-        let r = SRational { num: -10, denom: -20 };
+        let r = SRational::from((-10, -20));
         assert_eq!(format!("{}", r),         "-10/-20");
         assert_eq!(format!("{:11}", r),      "    -10/-20");
         assert_eq!(format!("{:3}", r),       "-10/-20");
@@ -750,32 +974,30 @@ mod tests {
     #[test]
     fn ratioanl_f64() {
         use std::{f64, u32};
-        assert_eq!(f64::from(Rational { num: 1, denom: 2 }), 0.5);
-        assert_eq!(f64::from(Rational { num: 1, denom: u32::MAX }),
+        assert_eq!(f64::from(Rational::from((1, 2))), 0.5);
+        assert_eq!(f64::from(Rational::from((1, u32::MAX))),
                    2.3283064370807974e-10);
-        assert_eq!(f64::from(Rational { num: u32::MAX, denom: 1 }),
+        assert_eq!(f64::from(Rational::from((u32::MAX, 1))),
                    u32::MAX as f64);
-        assert_eq!(f64::from(Rational { num: u32::MAX - 1, denom: u32::MAX }),
+        assert_eq!(f64::from(Rational::from((u32::MAX - 1, u32::MAX))),
                    0.9999999997671694);
-        assert_eq!(f64::from(Rational { num: u32::MAX, denom: u32::MAX - 1 }),
+        assert_eq!(f64::from(Rational::from((u32::MAX, u32::MAX - 1))),
                    1.0000000002328306);
-        assert_eq!(f64::from(Rational { num: 1, denom: 0 }), f64::INFINITY);
-        assert!(f64::from(Rational { num: 0, denom: 0 }).is_nan());
+        assert_eq!(f64::from(Rational::from((1, 0))), f64::INFINITY);
+        assert!(f64::from(Rational::from((0, 0))).is_nan());
 
-        assert_eq!(f64::from(SRational { num: 1, denom: 2 }), 0.5);
-        assert_eq!(f64::from(SRational { num: -1, denom: 2 }), -0.5);
-        assert_eq!(f64::from(SRational { num: 1, denom: -2 }), -0.5);
-        assert_eq!(f64::from(SRational { num: -1, denom: -2 }), 0.5);
-        assert_eq!(f64::from(SRational { num: 1, denom: 0 }), f64::INFINITY);
-        assert_eq!(f64::from(SRational { num: -1, denom: 0 }),
-                   f64::NEG_INFINITY);
+        assert_eq!(f64::from(SRational::from((1, 2))), 0.5);
+        assert_eq!(f64::from(SRational::from((-1, 2))), -0.5);
+        assert_eq!(f64::from(SRational::from((1, -2))), -0.5);
+        assert_eq!(f64::from(SRational::from((-1, -2))), 0.5);
+        assert_eq!(f64::from(SRational::from((1, 0))), f64::INFINITY);
+        assert_eq!(f64::from(SRational::from((-1, 0))), f64::NEG_INFINITY);
     }
 
     #[test]
     fn rational_f32() {
         // If num and demon are converted to f32 before the division,
         // the precision is lost in this example.
-        assert_eq!(f32::from(Rational { num: 1, denom: 16777217 }),
-                   5.960464e-8);
+        assert_eq!(f32::from(Rational::from((1, 16777217))), 5.960464e-8);
     }
 }
